/* msb.c: (C) TOYODA Eizi 1994
 * for NECK Project
 * character handler called when MSB is on.
 */

#include <stdio.h>
#include "neck.h"

#define	SJISHMAX	(0xDF)
#define	SJISAMAX	(0xEB)
#define	SJISBMIN	(0x40)
#define	SJISBMAX	(0xFC)

	static void
Unbelievable()
{
	flags.sj = SJIS_DEFAULT;
}

/* msb_sjis():
 * whether G1+C1 be taken as SJIS or EUC(JIS X 0202)
 */
	void
msb_sjis(arg)
	boolean		arg;
{
	flags.sj = arg;
}

#define	TurnToSjis()	(msb_sjis(MYTRUE))
#define	TurnToEuc()		(msb_sjis(MYFALSE))

	static void
SjisPret(ca, cb)
	int	ca, cb;
{
	int hi, lo;
	if (SJISHMAX < ca)
		ca = ca - 0x40;
	hi = 2 * (ca - 0x81) + 0x21;
	if (0x9E < cb) {
		hi++;
		lo = cb - 0x9F + 0x21;
	} else {
		if (0x7F < cb)
			cb--;
		lo = cb - 0x40 + 0x21;
	}
	myputwc(hi, lo);
}

	static void
MustbeSjis(ca)
	int	ca;
{
	int	cb;
	cb = mygetc(MYFALSE);
	if (cb < SJISBMIN) {
		Unbelievable();
		myputc(cb);
		return;
	}
	if (SJISBMAX < cb) {
		Unbelievable();
		return;
	}
	SjisPret(ca, cb);
}

	static void
MaybeEuc0201Katakana()
{
	int 	cb;
	cb = mygetc(MYFALSE);
	if (cb <= SPC) {
		Unbelievable();
		myputc(cb);
	} else if (M_SPC < cb && cb <= SJISBMAX) {
		/* that was SJIS */
		TurnToSjis();
		SjisPret(SS2, cb);
	} else if (SJISBMAX < cb) {
		/* 8bit graph upper */
		Unbelievable();
	} else {
		/* EUC 0201 katakana */
		myputc(cb);
	}
}

	static void
ButCantEuc(ca, cb)
	int	ca, cb;
{
	if (SJISAMAX < ca) {
		Unbelievable();
		(void)myungetc(cb);
	} else if (SJISHMAX < ca) {
		if (cb < SJISBMIN) {
			Unbelievable();
		} else {
			SjisPret(ca, cb);
		}
	} else {
		/* Sjis 1byte KATAKANA */
		myputc(ca);
		(void)myungetc(cb);
	}
}

	static void
MaybeEuc(ca)
	int	ca;
{
	int	cb;
	if ((cb = mygetc(MYTRUE)) == EOF) {
		/* take as SJIS 1byte Katakana */
		myputc(ca);
	} else if (cb <= M_SPC) {
		/* cannot EUC */
		TurnToSjis();
		ButCantEuc(ca, cb);
 	} else if (M_DEL == cb) {
		/* sjis 0201 katakana? */
		Unbelievable();
		myputc(ca);
		myputc(DEL);
	} else {
		/* true EUC */
		myputwc(ca & 0x7F, cb & 0x7F);
	}
}

	static void
MaybeSjis0201Katakana(ca)
	int	ca;
{
	int	cb;
	/* for anti-0201-katakana user */
	if (flags.k) {
		TurnToEuc();
		MaybeEuc(ca);
		return;
	}
	if ((cb = mygetc(MYTRUE)) == EOF) {
		/* take as SJIS 0201 katakana */
		myputc(ca);
	} else if (SJISBMAX < cb && cb < M_DEL) {
		/* must be EUC */
		TurnToEuc();
		myputwc(ca & 0x7F, cb & 0x7F);
	} else {
		/* normal SJIS 0201 katakana */
		myputc(ca);
		(void)myungetc(cb);
	}
}

	static void
MaybeSjis(ca)
	int	ca;
{
	if (ca <= SJISHMAX) {
		MaybeSjis0201Katakana(ca);
	} else if (SJISAMAX < ca) {
		TurnToEuc();
		MaybeEuc(ca);
	} else { /* --- maybe higher SJIS 2byte */
		int	cb;
		cb = mygetc(MYFALSE);
		if (cb < SJISBMIN || cb == M_DEL) {
			Unbelievable();
		} else if (SJISBMAX < cb) {
			/* ZITUHA EUC DE ATTA */
			TurnToEuc();
			myputwc(ca & 0x7F, cb & 0x7F);
		} else {
			SjisPret(ca, cb);
		}
	}
}

	void
msb_on(c)
	int	c;
{
	if (c == SS2 && !flags.sj) {
		MaybeEuc0201Katakana();
	} else if (c == CSI && (!flags.sj || flags.c)) {
		ansi_csi();
	} else if (c == M_SPC) {
		myputc(SPC);
	} else if (c == M_DEL) {
		myputc(DEL);
	} else if (c == 0x80) {
		myputc(0x80);
	} else if (0x80 < c && c < M_SPC) {
		TurnToSjis();
		MustbeSjis(c);
	} else if (!flags.sj) {
		/* --- M_SPC < c && c < M_DEL
		 * SJIS: 0201 katakana or SJIS-2byte
		 * EUC: EUC 2byte
		 */
		MaybeEuc(c);
	} else {
		MaybeSjis(c);
	}
}
